import argparse
import asyncio
import inspect
import logging
import logging.handlers
import os
from datetime import timezone
from typing import List

from kekmonitors.config import ERRORS, Config, LogConfig
from kekmonitors.utils.server.msg import Cmd, Response, badResponse, okResponse


def get_logger(config: LogConfig):
	'''Get preconfigured logger'''
	logger = logging.getLogger(config.name)
	logger.propagate = False
	logger.setLevel(logging.DEBUG)
	formatter = logging.Formatter(
		"[%(asctime)s] %(levelname)s: %(message)s")

	while logger.handlers:
		logger.handlers.pop()

	splitted_name = config.name.split(".")
	log_path = Config().log_path
	os.makedirs(os.path.sep.join(
		[log_path, *splitted_name[:2]]), exist_ok=True)
	file_handler = logging.handlers.TimedRotatingFileHandler(filename=os.path.sep.join(
		[log_path, *splitted_name[:2], "".join([splitted_name[-1], ".log"])]), when="midnight", interval=1, backupCount=7)
	file_handler.setLevel(config.file_level)
	file_handler.setFormatter(formatter)

	logger.addHandler(file_handler)

	if config.add_stream_handler:
		stream_handler = logging.StreamHandler()
		stream_handler.setLevel(config.stream_level)
		stream_handler.setFormatter(formatter)
		logger.addHandler(stream_handler)

	return logger


# https://stackoverflow.com/questions/31818050/round-number-to-nearest-integer
def proper_round(num, dec=0):
	num = str(num)[:str(num).index('.') + dec + 2]
	if num[-1] >= '5':
		return int(float(num[:-2 - (not dec)] + str(int(num[-2 - (not dec)]) + 1)))
	return int(float(num[:-1]))


def is_in_whitelist(full_name, whitelist, separator=","):
	'''For each string in whitelist, separate the string using separator and check if all splitted items are contained in full_name.\n
	This way for example you can check if either `air jordan 13` and/or `air high jordan 13` are in the whitelist=['air, jordan 13']'''
	if whitelist == []:
		return True
	for shoe_whitelist in whitelist:
		all_words_in_name = True
		for word_whitelist in shoe_whitelist.split(separator):
			if full_name.lower().find(word_whitelist) == -1:
				all_words_in_name = False
				break
		if all_words_in_name:
			return True

	return False


def utc_to_local(utc_dt):
	'''https://stackoverflow.com/questions/4563272/convert-a-python-utc-datetime-to-a-local-datetime-using-only-python-standard-lib'''
	return utc_dt.replace(tzinfo=timezone.utc).astimezone(tz=None)


def chunks(lst, n):
	"""https://stackoverflow.com/questions/312443/how-do-you-split-a-list-into-evenly-sized-chunks?noredirect=1&lq=1\n
	Yield successive n-sized chunks from lst."""
	for i in range(0, len(lst), n):
		yield lst[i:i + n]


def make_default_executable(_class, config: Config = Config()):
	parser = argparse.ArgumentParser(
            description=f"Default executable for {_class.__name__}, generated from utils.tools.make_default_executable")
	parser.add_argument("-d", "--delay", default=config.loop_delay, type=int,
                     help=f"Specify a delay for the loop. (default: {config.loop_delay})")
	parser.add_argument("--output", action=argparse.BooleanOptionalAction,
                     default=True,
                     help="Specify wether you want log output to the console or not. (note: this does not disable file log)",)
	args = parser.parse_args()
	if args.delay < 0:
		print(f"Cannot have a negative delay")
		return
	config.add_stream_handler = args.output
	config.loop_delay = args.delay
	_class(config).start(args.delay)


def dump_error(logger: logging.Logger, response: Response):
	e = response.error.value
	if e:
		curframe = inspect.currentframe()
		calframe = inspect.getouterframes(curframe, 2)
		fn_name = calframe[1][3]
		log = f"{fn_name}: Error: {response.error.name}"
		if response.info:
			log += f"; info: {response.info}"
		if response.payload:
			log += f"; payload: {response.payload}"
		logger.warning(log)


async def make_request(socket_path: str, cmd: Cmd, expect_response=True) -> Response:
	if os.path.exists(socket_path):
		try:
			reader, writer = await asyncio.open_unix_connection(socket_path)
			writer.write(cmd.get_bytes())
			writer.write_eof()

			if expect_response:
				response = Response(await reader.read())

				writer.close()
				return response
			return okResponse()
		except ConnectionRefusedError:
			pass
	r = badResponse()
	r.error = ERRORS.SOCKET_DOESNT_EXIST
	return r


def list_contains_find_item(l: List[str], s: str):
	for item in l:
		if item.find(s) != -1:
			return True
	else:
		False


def get_file_if_exist_else_create(filename_path, content) -> str:
	filename_directory_path = filename_path[:filename_path.rfind(os.path.sep)]
	os.makedirs(filename_directory_path, exist_ok=True)
	if os.path.isfile(filename_path):
		with open(filename_path, "r") as rf:
			return rf.read()
	else:
		with open(filename_path, "w") as wf:
			wf.write(content)
		return content



# This is an incomplete web app to access the monitor manager from the browser

import json

import tornado.ioloop
import tornado.web

from kekmonitors.config import COMMANDS, Config
from kekmonitors.utils.server.msg import Cmd
from kekmonitors.utils.tools import make_request

config = Config()


async def send_to_moman(cmd: Cmd):
	'''Send a command to the monitor manager'''
	return await make_request(f"{config.socket_path}/MonitorManager", cmd)


class RootHandler(tornado.web.RequestHandler):
	async def get(self):
		self.write(
			{
				"available endpoints:": list(ep[0] for ep in endpoints)
			}
		)


class AddMonitorHandler(tornado.web.RequestHandler):
	async def post(self):
		cmd = Cmd()
		cmd.cmd = COMMANDS.MM_ADD_MONITOR
		cmd.payload = json.loads(self.request.body)
		r = await send_to_moman(cmd)

		if r.error.value:
			self.set_status(400, r.reason)
		self.write(r.get_json())


class AddScraperHandler(tornado.web.RequestHandler):
	async def post(self):
		cmd = Cmd()
		cmd.cmd = COMMANDS.MM_ADD_SCRAPER
		cmd.payload = json.loads(self.request.body)
		r = await send_to_moman(cmd)

		if not r.error.value:
			self.set_status(400, r.reason)
		self.write(r.get_json())


class AddHandler(tornado.web.RequestHandler):
	async def post(self):
		cmd = Cmd()
		cmd.cmd = COMMANDS.MM_ADD_MONITOR_SCRAPER
		cmd.payload = json.loads(self.request.body)
		r = await send_to_moman(cmd)

		if not r.error.value:
			self.set_status(400, r.reason)
		self.write(r.get_json())


class StopMonitorHandler(tornado.web.RequestHandler):
	async def delete(self):
		cmd = Cmd()
		cmd.cmd = COMMANDS.MM_STOP_MONITOR
		cmd.payload = json.loads(self.request.body)
		r = await send_to_moman(cmd)

		if not r.error.value:
			self.set_status(400, r.reason)
		self.write(r.get_json())


class StopScraperHandler(tornado.web.RequestHandler):
	async def delete(self):
		cmd = Cmd()
		cmd.cmd = COMMANDS.MM_STOP_SCRAPER
		cmd.payload = json.loads(self.request.body)
		r = await send_to_moman(cmd)

		if not r.error.value:
			self.set_status(400, r.reason)
		self.write(r.get_json())


class StopHandler(tornado.web.RequestHandler):
	async def delete(self):
		cmd = Cmd()
		cmd.cmd = COMMANDS.MM_STOP_MONITOR_SCRAPER
		cmd.payload = json.loads(self.request.body)
		r = await send_to_moman(cmd)

		if not r.error.value:
			self.set_status(400, r.reason)
		self.write(r.get_json())


class MonitorStatusHandler(tornado.web.RequestHandler):
	async def get(self):
		cmd = Cmd()
		cmd.cmd = COMMANDS.MM_GET_MONITOR_STATUS
		r = await send_to_moman(cmd)
		self.write(r.get_json())


class ScraperStatusHandler(tornado.web.RequestHandler):
	async def get(self):
		cmd = Cmd()
		cmd.cmd = COMMANDS.MM_GET_SCRAPER_STATUS
		r = await send_to_moman(cmd)
		self.write(r.get_json())


class StatusHandler(tornado.web.RequestHandler):
	async def get(self):
		cmd = Cmd()
		cmd.cmd = COMMANDS.MM_GET_MONITOR_SCRAPER_STATUS
		r = await send_to_moman(cmd)
		self.write(r.get_json())


# add endpoints here
endpoints = [
	("/", RootHandler),
	("/monitors/add", AddMonitorHandler),
	("/scrapers/add", AddScraperHandler),
	("/add", AddHandler),
	("/monitors/stop", StopMonitorHandler),
	("/scrapers/stop", StopScraperHandler),
	("/stop", StopHandler),
	("/monitors/status", MonitorStatusHandler),
	("/scrapers/status", ScraperStatusHandler),
	("/status", StatusHandler),
]

if __name__ == "__main__":
	app = tornado.web.Application(endpoints)
	app.listen(8888)

	tornado.ioloop.IOLoop.current().start()



from datetime import datetime

from discord import Embed
from kekmonitors.utils import shoe_stuff
from kekmonitors.utils.shoe_stuff import Shoe


def get_empty_embed() -> Embed:
	'''Get an almost empty embed (only set color and timestamp)'''
	empty_embed = Embed()
	empty_embed.timestamp = datetime.utcnow()
	return empty_embed


def add_link(s: str, link: str) -> str:
	'''Add a link to the string s in markdown format'''
	return "[" + s + "](" + link + ")"


def get_default_embed(shoe: Shoe, allow_unavailable_sizes=False, show_website=False):
	'''Do you really need help to understand this?'''
	embed = get_empty_embed()

	embed.title = shoe.name
	embed.url = shoe.link
	embed.set_thumbnail(url=shoe.img_link)

	if show_website:
		if not isinstance(embed.description, str):
			embed.description = ""
		embed.description += "\n" + \
			shoe.link[shoe.link.find("/") + 2:shoe.link.find("/", 8)]

	if shoe.reason == shoe_stuff.NEW_RELEASE:
		embed.add_field(name="Notification type", value="New item", inline=False)
	if shoe.reason == shoe_stuff.RESTOCK:
		embed.add_field(name="Notification type", value="Restock", inline=False)

	if shoe.price != "Not available":
		embed.add_field(name="Price", value=shoe.price, inline=False)

	if shoe.release_date != "":
		if shoe.release_method != "":
			embed.add_field(name="Release date", value=shoe.release_date)
			embed.add_field(name="Release method", value=shoe.release_method)
		else:
			embed.add_field(name="Release date", value=shoe.release_date)

	# Add sizes to the embed
	sizes = []
	total_chars = 0
	for size in shoe.sizes:
		available = shoe.sizes[size]["available"]
		# Only add sizes under these conditions
		if available or (not available and allow_unavailable_sizes):
			size_str = size
			# Add the stock number if available
			if "stock" in shoe.sizes[size]:
				size_str += " (" + str(shoe.sizes[size]["stock"]) + ")"
			# Add atc if available
			if "atc" in shoe.sizes[size] and shoe.sizes[size]["atc"] != "":
				atc = shoe.sizes[size]["atc"]
				atc_str = add_link("[ATC]", atc)
				size_str += " - " + atc_str
			# Add quick tasks if available
			if "quick_tasks" in shoe.sizes[size]:
				qts = []
				for key, value in zip(shoe.sizes[size]["quick_tasks"].keys(), shoe.sizes[size]["quick_tasks"].values()):
					qts.append(add_link(key, value))
				qts_str = " - ".join(qts)
				size_str += " - " + qts_str
			# Add anything else
			if "other" in shoe.sizes[size]:
				others_list = []
				for other in shoe.sizes[size]["other"]:
					other_str = ""
					if "link" in other:
						other_str = add_link(other["name"], other["link"])
					else:
						other_str = other["name"]
					others_list.append(other_str)
				others_str = " - ".join(others_list)
				size_str += " - " + others_str
			sizes.append(size_str)
			total_chars += len(size_str)

	# get well formatted sizes.
	values = get_valid_values(sizes, 6)

	for index, value in enumerate(values):
		if index == 0:
			inline = True if len(values) > 0 else False
			embed.add_field(name="Sizes", value=value, inline=inline)
		else:
			embed.add_field(name="\u200b", value=value, inline=True)

	return embed


def get_valid_values(values_list, max_elements):
	'''This function tries to format a list splitting it in max_elements and keeping count of discord's max chars limit.\n
		See comments for more insight.'''
	if len(values_list) == 0:
		return []
	# not perfect. if many small values and very close to 1023 chars it might truncate the lasts of them.
	# also, this requires too much fucking math.
	tmp = ""
	done = False
	valid_values = []
	count = 0
	for value in values_list:
		done = False
		if len(tmp) + len("\n") + len(value) > 1023 or (count != 0 and count % max_elements == 0):
			if tmp == "":
				# first cycle / 1 item
				tmp = value
			done = True
			valid_values.append(tmp)
			tmp = ""
			count = 0
		tmp += value + "\n"
		count += 1
	if not done:
		valid_values.append(tmp)
	return valid_values


def website_specific_embed(shoe: Shoe) -> Embed:
	'''What a website-specific embed function might look like.'''
	fd_embed = get_default_embed(shoe)
	if shoe.reason == shoe_stuff.NEW_RELEASE:
		if shoe.sizes == {}:
			fd_embed.add_field(
				name="Info", value="If sizes are not displayed check manually!", inline=False)
	if "alternative_url" in shoe.tags:
		fd_embed.description = f"\n\n[Alternative link]({shoe.tags['alternative_url']})"
	return fd_embed



import asyncio
from datetime import datetime
from typing import Dict, Optional, Tuple

import pycurl
import tornado.httpclient
from kekmonitors.config import LogConfig
from kekmonitors.utils.tools import get_logger
from tornado.curl_httpclient import CurlError


class NetworkUtils(object):
	def __init__(self, logger_name: str):
		self.asyncio_loop = asyncio.get_event_loop()
		tornado.httpclient.AsyncHTTPClient.configure(
			"tornado.curl_httpclient.CurlAsyncHTTPClient")
		self.client = tornado.httpclient.AsyncHTTPClient()
		self._last_modified_datetimes = {}  # type: Dict[str, datetime]
		# force a cache refresh after self.cache_timeout
		self.cache_timeout = 10 * 60
		# keep a copy of every page in memory, for when we receive 304
		self._cached_pages = {}  # type: Dict[str, str]
		# remember etags, sometimes some websites use them
		self._etags = {}  # type: Dict[str, str]

		for feature in pycurl.version.split(" "):
			if feature.find("brotli") != -1:
				self._has_brotli = True
				break
		else:
			self._has_brotli = False

		config = LogConfig()
		config.name = f"{logger_name}.NetworkUtils"
		self.network_logger = get_logger(config)
		self.network_logger.debug(f"Has brotli: {self._has_brotli}")

	async def fetch(self, url: str, use_cache=True, attempts=3, delay=2, *args, **kwargs) -> Tuple[Optional[tornado.httpclient.HTTPResponse], str]:
		'''Asynchronously fetch the url using a tornado client. If you want to fetch more urls at once use asyncio.gather(*tasks).\n
		You can pass arguments to client.fetch() using *args and **kwargs (e.g. if you need proxies you can call self.fetch like this:\n
		`self.fetch(url, headers=headers, proxy_host={your-proxy-host}, proxy_port={your-proxy-port})`'''
		total_attempts = attempts
		headers = kwargs.get("headers", {})
		# keep retrying the connection until we run out of attempts
		while attempts > 0:
			try:
				if_mod_since = None
				# if using cache and it has expired/timed out
				if use_cache and url in self._cached_pages and url in self._last_modified_datetimes:
					if (datetime.utcnow() - self._last_modified_datetimes[url]).seconds < self.cache_timeout:
						self.network_logger.debug(url + " is in cache, adding cache headers")
						if_mod_since = self._last_modified_datetimes[url]
						if url in self._etags:
							headers["if-none-match"] = self._etags[url]
					else:
						self.network_logger.info(url + " is in cache from more than " +
                                                    str(self.cache_timeout) + " seconds, refreshing the page.")
						self._cached_pages.pop(url)
						self._last_modified_datetimes.pop(url)

				# fix some possibly set headers from fake-headers
				headers["accept-encoding"] = "gzip, deflate"
				if self._has_brotli:
					headers["accept-encoding"] += ", br"
				headers.pop("pragma", None)
				headers.pop("Pragma", None)

				self.network_logger.debug(f"Getting {url}...")
				r = await self.client.fetch(url, if_modified_since=if_mod_since, raise_error=False, *args, **kwargs)
				self.network_logger.info(f"Got {url} with code {r.code}")
				if r.code == 599:
					self.network_logger.warning(
						f"Something happened while fetching {url}: {r.reason}")
					attempts -= 1
					if attempts:
						# sleep delay before retrying
						await asyncio.sleep(delay)
						continue
				elif r.code == 404:
					# usually this is what we want to do
					self.network_logger.debug("Not retrying on 404.")
					return r, r.body.decode()
				else:
					if use_cache:
						# if page was cached update it or just return it
						if r.code < 400 and r.code != 304:
							self._cached_pages[url] = r.body.decode()
							self._last_modified_datetimes[url] = datetime.utcnow()
							if "etag" in r.headers:
								self._etags[url] = r.headers["etag"]
						return r, self._cached_pages[url]
					else:
						return r, r.body.decode()

			except CurlError:
				self.network_logger.exception(f"Timed out while fetching {url}.")
				attempts -= 1
				if attempts:
					await asyncio.sleep(delay)
					continue

			except:
				self.network_logger.exception("Got exception:")
				await asyncio.sleep(delay)
				attempts -= 1
				if attempts:
					await asyncio.sleep(delay)
					continue

		self.network_logger.warning(
			f"Tried {total_attempts} but couldn't fetch {url}.")

		return self.__dict__.get("r", None), ""



from typing import Any, Dict, List

import pymongo
from kekmonitors.config import Config
from kekmonitors.utils.shoe_stuff import Shoe


def _change_keys(d, _from, _to):
	if isinstance(d, list):
		new_list = []
		for element in d:
			new_list.append(_change_keys(element, _from, _to))
		return new_list
	elif isinstance(d, dict):
		new_dict = {}
		for key, value in d.items():
			value = _change_keys(value, _from, _to)
			new_key = key
			if isinstance(key, str) and _from in key:
				new_key = key.replace(_from, _to)
			new_dict[new_key] = value
		return new_dict
	else:
		return d


def sanitize(d: Dict[str, Any]) -> Dict[str, Any]:
	"""
	Converts keys with "." to "_dot_", as it's not possile to add keys with a "." in MongoDBc
	"""
	return _change_keys(d, ".", "_dot_")


def unsanitize(d: Dict[str, Any]) -> Dict[str, Any]:
	"""
	Converts keys with "_dot_" back to "."
	"""
	return _change_keys(d, "_dot_", ".")


class ShoeManager(object):
	'''Manages the database. Mostly a MongoDB wrapper.'''

	def __init__(self):
		config = Config()
		self.db_name = config.db_name
		self.db_path = config.db_path

		self._db = pymongo.MongoClient(self.db_path)[self.db_name]["items"]

	def add_shoe(self, shoe: Shoe):
		'''Add this shoe to the database'''
		self._db.insert_one(sanitize(shoe.__dict__))

	def add_shoes(self, shoes: List[Shoe]):
		'''Add these shoes to the database'''
		l = []
		for shoe in shoes:
			l.append(sanitize(shoe.__dict__))
		self._db.insert_many(l)

	def find_shoe(self, query: Dict[str, Any]):
		'''Find a shoe passing ```query``` to pymongo's find_one'''
		q = {}
		for k in query:
			if not k.startswith("_Shoe__"):
				q[f"_Shoe__{k}"] = query[k]
			else:
				q[k] = query[k]
		item = self._db.find_one(q, {"_id": 0})
		if item:
			shoe = Shoe()
			shoe.__dict__ = unsanitize(item)
			return shoe
		else:
			return None

	def find_shoes(self, query: Dict[str, Any]):
		'''Find all shoes passing ```query``` to pymongo's find_one'''
		q = {}
		for k in query:
			if not k.startswith("_Shoe__"):
				q[f"_Shoe__{k}"] = query[k]
			else:
				q[k] = query[k]
		items = self._db.find(q, {"_id": 0})
		shoes = []
		for item in items:
			if item:
				shoe = Shoe()
				shoe.__dict__ = unsanitize(item)
				shoes.append(shoe)
		return shoes

	def update_shoe(self, shoe: Shoe):
		'''Update the shoe in the db matching the same link.'''
		self._db.update_many({"_Shoe__link": shoe.link}, {
		                     "$set": sanitize(shoe.__dict__)})



import json
import time
from datetime import datetime
from logging import Logger
from queue import Queue
from threading import Event, Thread
from typing import Any, Dict, List, Tuple

import requests
from discord import Embed
from kekmonitors.config import Config, LogConfig
from kekmonitors.utils.tools import get_logger


class WebhookSender(Thread):
	'''This handles sending embeds to one specific webhook. Obviously being a thread sub-class you should not add too many webhooks (in the order of the hundreds) for the same monitor.\n
	You should not use this directly, but `WebhookManager` instead'''

	def __init__(self, webhook: str, config: Config):
		self.config = config
		self.webhook = webhook
		self.logger = Logger(self.config.name)
		self.queue = Queue()  # type: Queue[Tuple[List[Any], List[Any], datetime]]
		# contains the webhook config, embeds and time at which they were added
		self.add_event = Event()
		self.has_to_quit = False
		super().__init__()

	def add_to_queue(self, webhook_values: Any, embed: Embed):
		now = datetime.now()
		self.queue.put((webhook_values, embed, now))
		self.add_event.set()

	def is_done(self) -> bool:
		return not self.add_event.isSet()

	def quit(self):
		self.has_to_quit = True
		self.add_event.set()

	def run(self):
		while True:
			self.add_event.wait()
			if self.has_to_quit:
				break
			while not self.queue.empty():
				webhook_values, embed, now = self.queue.get()
				if "custom" in webhook_values:
					provider = webhook_values["custom"].get(
						"provider", self.config.provider)
					timestamp_format = webhook_values["custom"].get(
						"timestamp_format", self.config.timestamp_format)
					ts = now.strftime(timestamp_format)
					icon_url = webhook_values["custom"].get(
						"icon_url", self.self.config.provider_icon)
					color = webhook_values["custom"].get(
						"color", self.self.config.embed_color)

					embed.set_footer(text=" | ".join([provider, ts]), icon_url=icon_url)
					embed.color = color
				else:
					ts = now.strftime(self.config.timestamp_format)

					embed.set_footer(text=f"{self.config.provider} | {ts}",
					                 icon_url=self.self.config.provider_icon)
					embed.color = self.self.config.embed_color

				embed.timestamp = Embed.Empty
				data = {"embeds": [embed.to_dict()]}

				if "custom" in webhook_values:
					if "avatar_image" in webhook_values["custom"]:
						data["avatar_url"] = webhook_values["custom"]["avatar_image"]

				data = json.dumps(data)

				while True:
					r = requests.post(self.webhook, data=data, headers={
					                  "Content-Type": "application/json"}, timeout=3)
					self.logger.debug(f"Posted to {self.webhook} with code {r.status_code}")
					remaining_requests = r.headers["x-rateLimit-remaining"]
					if remaining_requests == "0":
						delay = int(r.headers["x-rateLimit-reset-after"])
						self.logger.debug(
							f"No available requests reminaing for {self.webhook}, waiting {delay} secs")
						time.sleep(delay)
					if r.status_code == 429:
						delay = r.headers["x-rateLimit-reset-after"]
						self.logger.debug(f"Got 429 for {self.webhook}, waiting {delay} secs")
						time.sleep(delay)
						continue
					break
			self.add_event.clear()


class WebhookManager():
	def __init__(self, config: Config):
		self.config = config
		logconfig = LogConfig(self.config)
		logconfig.name += ".WebhookManager"
		self.logger = get_logger(logconfig)
		self.webhook_senders = {}  # type: Dict[str, WebhookSender]
		self.add_event = Event()
		self.logger.debug("Started webhook manager")

	def quit(self):
		self.logger.debug("Starting shutdown...")
		for w in self.webhook_senders:
			ws = self.webhook_senders[w]
			while not ws.is_done():
				time.sleep(0.5)
			ws.quit()

		for w in self.webhook_senders:
			ws = self.webhook_senders[w]
			ws.join()
		self.logger.debug("Shut down...")

	def add_to_queue(self, embed: Embed, webhooks: Dict[str, Dict[str, Any]]):
		'''Add the embed to the queue of webhooks to send. It will be processed as soon as possible.'''
		for webhook in webhooks:
			if webhook in self.webhook_senders:
				self.webhook_senders[webhook].add_to_queue(webhooks[webhook], embed)
			else:
				self.webhook_senders[webhook] = WebhookSender(webhook, self.config)
				self.webhook_senders[webhook].start()
				self.webhook_senders[webhook].add_to_queue(webhooks[webhook], embed)



import pymongo
from kekmonitors.config import Config

if __name__ == "__main__":
	config = Config()
	c = input(
		f"Executing this file will destroy {config.db_name} (ALL WEBSITES, MONITORS, SCRAPERS) unrecoverably. Are you sure you want to proceed? (y/n) ")
	if c != "y":
		print("Exiting (no modifications have been made.)")
		exit(0)
	c = input("Are you sure? Last chance. (y/n) ")
	if c != "y":
		print("Exiting (no modifications have been made.)")
		exit(0)

	root_db = pymongo.MongoClient(config.db_path)
	db = root_db[config.db_name]
	db["items"].drop()
	db["register"]["Monitors"].drop()
	db["register"]["Scrapers"].drop()

	print("Done!")



import asyncio
import json
import os
from json.decoder import JSONDecodeError
from typing import Any, Dict, List, Optional

import __main__
import pymongo
from pymongo.collection import Collection

from kekmonitors.config import COMMANDS, ERRORS, Config, LogConfig
from kekmonitors.utils.server.msg import Cmd, Response, badResponse, okResponse
from kekmonitors.utils.server.server import Server
from kekmonitors.utils.tools import get_file_if_exist_else_create, get_logger


def mark_as(_type: str, name: str, path: str, client: Collection):
	existing = client[_type].find_one({"name": name})
	if not existing:
		client[_type].insert_one({"name": name, "path": path})
	else:
		if existing["path"] != path:
			raise Exception(
				f"Trying to register new {_type} ({name}) when it already exists in the database with a different path: {existing['path']}")


class Common(Server):
	def __init__(self, config: Config):
		self.config = config

		log_config = LogConfig(config)

		log_config.name += ".General"
		self.general_logger = get_logger(log_config)
		log_config.name = self.config.name + ".Client"
		self.client_logger = get_logger(log_config)

		self.class_name = self.get_class_name()

		super().__init__(config, os.path.sep.join(
			[self.config.socket_path, config.name]))

		self.db_client = pymongo.MongoClient(self.config.db_path)[
                    self.config.db_name]["register"]

		self._loop_lock = asyncio.Lock()

		self.cmd_to_callback[COMMANDS.SET_WHITELIST] = self.on_set_whitelist
		self.cmd_to_callback[COMMANDS.SET_BLACKLIST] = self.on_set_blacklist
		self.cmd_to_callback[COMMANDS.SET_WEBHOOKS] = self.on_set_webhooks
		self.cmd_to_callback[COMMANDS.SET_CONFIG] = self.on_set_config
		self.cmd_to_callback[COMMANDS.GET_WHITELIST] = self.on_get_whitelist
		self.cmd_to_callback[COMMANDS.GET_BLACKLIST] = self.on_get_blacklist
		self.cmd_to_callback[COMMANDS.GET_WEBHOOKS] = self.on_get_webhooks
		self.cmd_to_callback[COMMANDS.GET_CONFIG] = self.on_get_config

		is_monitor = config.name.startswith("Monitor.")
		pre_conf_path = self.config.config_path
		self.whitelist_json_filepath = os.path.sep.join(
			[pre_conf_path, "monitors" if is_monitor else "scrapers", "whitelists.json"])
		self.blacklist_json_filepath = os.path.sep.join(
			[pre_conf_path, "monitors" if is_monitor else "scrapers", "blacklists.json"])
		self.webhooks_json_filepath = os.path.sep.join(
			[pre_conf_path, "monitors" if is_monitor else "scrapers", "webhooks.json"])
		self.config_json_filepath = os.path.sep.join(
			[pre_conf_path, "monitors" if is_monitor else "scrapers", "configs.json"])

		self.whitelist_json = self.load_config(self.whitelist_json_filepath)
		self.blacklist_json = self.load_config(self.blacklist_json_filepath)
		self.webhooks_json = self.load_config(self.webhooks_json_filepath)
		self.config_json = self.load_config(self.config_json_filepath)

		self._new_whitelist = None  # type: Optional[List[str]]
		self._new_blacklist = None  # type: Optional[List[str]]
		self._new_webhooks = None  # type: Optional[Dict[str, Dict[str, Any]]]
		self._new_config = None  # type: Optional[Dict[str, Any]]

		self._has_to_quit = False

	def init(self):
		'''Override this in your website-specific monitor, if needed.'''
		pass

	async def async_init(self):
		'''Override this in your website-specific monitor, if needed.'''
		pass

	def on_shutdown(self):
		'''Override this in your website-specific monitor, if needed.'''
		pass

	async def on_async_shutdown(self):
		'''Override this in your website-specific monitor, if needed.'''
		pass

	def get_class_name(self):
		'''Internal function used to get the correct filename.'''
		'''Not necessary, but sometimes you might want to override this.'''
		return type(self).__name__

	def _mark_as_monitor(self):
		mark_as("monitors", self.class_name, __main__.__file__, self.db_client)

	def _mark_as_scraper(self):
		mark_as("scrapers", self.class_name, __main__.__file__, self.db_client)

	def load_config(self, path):
		content = get_file_if_exist_else_create(path, "{}")
		try:
			j = json.loads(content)
		except JSONDecodeError:
			self.general_logger.exception(f"{path} is not valid json. Quitting.")
			exit(1)
		if self.class_name not in j:
			self.general_logger.warning(
				f"{self.class_name} not in {path}, continuing with empty entry.")
			return {}
		else:
			return j[self.class_name]

	def update_local_config(self):
		if self._new_blacklist is not None:
			self.general_logger.info(f"New blacklist: {self._new_blacklist}")
			self.blacklist_json = self._new_blacklist
			self._new_blacklist = None
		if self._new_whitelist is not None:
			self.general_logger.info(f"New whitelist: {self._new_whitelist}")
			self.whitelist_json = self._new_whitelist
			self._new_whitelist = None
		if self._new_webhooks is not None:
			self.general_logger.info(f"New webhooks: {self._new_webhooks}")
			self.webhooks_json = self._new_webhooks
			self._new_webhooks = None
		if self._new_config is not None:
			self.general_logger.info(f"New config: {self._new_config}")
			self.config_json = self._new_config
			self._new_config = None

	async def on_set_whitelist(self, cmd: Cmd) -> Response:
		r = badResponse()
		whitelist = cmd.payload
		if isinstance(whitelist, list):
			self._new_whitelist = whitelist
			self.client_logger.debug("Got new whitelist")
			r = okResponse()
		else:
			self.client_logger.warning(
				f"Got new whitelist but it was invalid: {whitelist}")
			r.error = ERRORS.BAD_PAYLOAD
			r.info = f"Invalid whitelist (expected list, got {type(cmd.payload)}"
		return r

	async def on_set_blacklist(self, cmd: Cmd) -> Response:
		r = badResponse()
		blacklist = cmd.payload
		if isinstance(blacklist, list):
			self._new_blacklist = blacklist
			self.client_logger.debug("Got new blacklist")
			r = okResponse()
		else:
			self.client_logger.warning(
				f"Got new blacklist but it was invalid: {blacklist}")
			r.error = ERRORS.BAD_PAYLOAD
			r.info = f"Invalid blacklist (expected list, got {type(cmd.payload)}"
		return r

	async def on_set_webhooks(self, cmd: Cmd) -> Response:
		r = badResponse()
		webhooks = cmd.payload
		if isinstance(webhooks, dict):
			self._new_webhooks = webhooks
			self.client_logger.debug("Got new webhooks")
			r = okResponse()
		else:
			self.client_logger.warning(
				f"Got new webhooks but it was invalid: {webhooks}")
			r.error = ERRORS.BAD_PAYLOAD
			r.info = f"Invalid webhooks (expected dict, got {type(cmd.payload)}"
		return r

	async def on_set_config(self, cmd: Cmd) -> Response:
		r = badResponse()
		config = cmd.payload
		if isinstance(config, dict):
			self._new_config = config
			self.client_logger.debug("Got new config")
			r = okResponse()
		else:
			self.client_logger.warning(
				f"Got new config but it was invalid: {config}")
			r.error = ERRORS.BAD_PAYLOAD
			r.info = f"Invalid config (expected dict, got {type(cmd.payload)}"
		return r

	async def on_get_config(self, cmd: Cmd) -> Response:
		r = okResponse()
		r.payload = self.config_json
		return r

	async def on_get_whitelist(self, cmd: Cmd) -> Response:
		r = okResponse()
		r.payload = self.whitelist_json
		return r

	async def on_get_blacklist(self, cmd: Cmd) -> Response:
		r = okResponse()
		r.payload = self.blacklist_json
		return r

	async def on_get_webhooks(self, cmd: Cmd) -> Response:
		r = okResponse()
		r.payload = self.webhooks_json
		return r

	async def main(self):
		pass

	def start(self, delay):
		'''Call this to start the loop.'''
		self.delay = delay
		self._main_task = self._asyncio_loop.create_task(self.main())
		self._asyncio_loop.run_forever()
		self.general_logger.debug("Shutting down...")



OTHER = 0
NEW_RELEASE = 1
RESTOCK = 2
INCOMING = 3


class Shoe:
	def __init__(self):
		self.link = ""
		self.img_link = ""
		self.name = ""
		self.style_code = ""
		self.price = "Not available"
		self.sizes = {}
		# self.sizes = {
		# 	"size": {
		#		"available": True, # only required key
		# 		"atc": "link",
		# 		"quick_tasks": {
		# 			"name1": "link1",
		# 			"name2": "link2"
		# 		},
		#		"other": [
		# 			{
		#				name: "name",
		#				link: "link"
		# 			},
		# 			{
		#				name: "name",
		#				link: "link"
		# 			}
		# 		]
		# 	}
		# }
		self.in_stock = False
		self.release_date = ""
		self.back_in_stock = False
		self.out_of_stock = False
		self.release_method = ""
		self.reason = OTHER
		self.tags = {}
		self.other = {}

	@property
	def link(self):
		return self.__link

	@link.setter
	def link(self, link):
		if not isinstance(link, str):
			raise Exception("Link is not a string")
		self.__link = link

	@property
	def img_link(self):
		return self.__img_link

	@img_link.setter
	def img_link(self, img_link):
		if not isinstance(img_link, str):
			raise Exception("Img_link is not a string")
		self.__img_link = img_link

	@property
	def name(self):
		return self.__name

	@name.setter
	def name(self, name):
		if not isinstance(name, str):
			raise Exception("Name is not a string")
		self.__name = name

	@property
	def style_code(self):
		return self.__style_code

	@style_code.setter
	def style_code(self, style_code):
		if not isinstance(style_code, str):
			raise Exception("Style_code is not a string")
		self.__style_code = style_code

	@property
	def price(self):
		return self.__price

	@price.setter
	def price(self, price):
		if not isinstance(price, str):
			raise Exception("Price is not a string")
		self.__price = price

	@property
	def sizes(self):
		return self.__sizes

	@sizes.setter
	def sizes(self, sizes):
		if not isinstance(sizes, dict):
			raise Exception("Sizes is not a dict")
		self.__sizes = sizes

	@property
	def in_stock(self):
		return self.__in_stock

	@in_stock.setter
	def in_stock(self, in_stock):
		if not isinstance(in_stock, bool):
			raise Exception("In_stock is not a bool")
		self.__in_stock = in_stock

	@property
	def out_of_stock(self):
		return self.__out_of_stock

	@out_of_stock.setter
	def out_of_stock(self, out_of_stock):
		if not isinstance(out_of_stock, bool):
			raise Exception("Out_of_stock is not a bool")
		self.__out_of_stock = out_of_stock

	@property
	def back_in_stock(self):
		return self.__back_in_stock

	@back_in_stock.setter
	def back_in_stock(self, back_in_stock):
		if not isinstance(back_in_stock, bool):
			raise Exception("Back_in_stock is not a bool")
		self.__back_in_stock = back_in_stock

	@property
	def release_date(self):
		return self.__release_date

	@release_date.setter
	def release_date(self, release_date):
		if not isinstance(release_date, str):
			raise Exception("Release_date is not a string")
		self.__release_date = release_date

	@property
	def release_method(self):
		return self.__release_method

	@release_method.setter
	def release_method(self, release_method):
		if not isinstance(release_method, str):
			raise Exception("Release_method is not a string")
		self.__release_method = release_method

	@property
	def reason(self):
		return self.__reason

	@reason.setter
	def reason(self, reason):
		if not isinstance(reason, int):
			raise Exception("Reason is not an int")
		if reason != OTHER and reason != NEW_RELEASE and reason != RESTOCK and reason != INCOMING:
			raise Exception("Invalid reason provided")
		self.__reason = reason
